import uuid
import copy
import logging
import contextlib
from datetime import datetime

import sirius_sdk
from sirius_sdk.agent.microledgers.abstract import AbstractBatchedAPI
from sirius_sdk.hub import CoProtocolThreadedTheirs, CoProtocolThreadedP2P
from sirius_sdk.base import AbstractStateMachine
from sirius_sdk.agent.aries_rfc.feature_0015_acks import Ack, Status
from sirius_sdk.agent.consensus.simple.messages import *
from sirius_sdk.agent.consensus import Locking

# Problem codes
REQUEST_NOT_ACCEPTED = "request_not_accepted"
REQUEST_PROCESSING_ERROR = 'request_processing_error'
RESPONSE_NOT_ACCEPTED = "response_not_accepted"
RESPONSE_PROCESSING_ERROR = 'response_processing_error'


class MicroLedgerSimpleConsensus(AbstractStateMachine):

    def __init__(self, me: Pairwise.Me, time_to_live: int = 60, logger=None, *args, **kwargs):
        super().__init__(time_to_live=time_to_live, logger=logger, *args, **kwargs)
        self.__me = me
        self.__problem_report = None
        self.__cached_p2p = {}

    @property
    def me(self) -> Pairwise.Me:
        return self.__me

    @property
    def problem_report(self) -> SimpleConsensusProblemReport:
        return self.__problem_report

    @contextlib.asynccontextmanager
    async def acceptors(self, theirs: List[Pairwise], thread_id: str):
        co = CoProtocolThreadedTheirs(
            thid=thread_id,
            theirs=theirs
        )
        self._register_for_aborting(co)
        try:
            try:
                yield co
            except OperationAbortedManually:
                await self.log(progress=100, message='Aborted')
                raise StateMachineAborted('Aborted by User')
        finally:
            self._unregister_for_aborting(co)

    @contextlib.asynccontextmanager
    async def leader(self, their: Pairwise, thread_id: str, time_to_live: int = None):
        co = CoProtocolThreadedP2P(
            thid=thread_id,
            to=their,
            time_to_live=time_to_live or self.time_to_live
        )
        self._register_for_aborting(co)
        try:
            try:
                yield co
            except OperationAbortedManually:
                await self.log(progress=100, message='Aborted')
                raise StateMachineAborted('Aborted by User')
        finally:
            self._unregister_for_aborting(co)

    async def init_microledger(
            self, ledger_name: str, participants: List[str], genesis: List[Transaction]
    ) -> (bool, AbstractMicroledger):
        """
        :param ledger_name: name of new microledger
        :param participants: list of DIDs that present pairwise list of the Microledger relationships
                (Assumed DIDs are public or every participant has relationship with each other via pairwise)
        :param genesis: genesis block of the new microledger if all participants accept transaction
        """
        await self._bootstrap(participants)
        relationships = [p2p for p2p in self.__cached_p2p.values()]
        async with self.acceptors(theirs=relationships, thread_id='simple-consensus-init-' + uuid.uuid4().hex) as co:
            await self.log(progress=0, message=f'Create ledger [{ledger_name}]')
            ledger, genesis = await sirius_sdk.Microledgers.create(ledger_name, genesis)
            await self.log(message=f'Ledger creation terminated successfully')
            try:
                await self._init_microledger_internal(co, ledger, participants, genesis)
                await self.log(progress=100, message='All participants accepted ledger creation')
            except Exception as e:
                await sirius_sdk.Microledgers.reset(ledger_name)
                await self.log(message=f'Reset ledger')
                if isinstance(e, StateMachineTerminatedWithError):
                    self.__problem_report = SimpleConsensusProblemReport(e.problem_code, e.explain)
                    await self.log(
                        progress=100, message=f'Terminated with error',
                        problem_code=e.problem_code, explain=e.explain
                    )
                    if e.notify:
                        await co.send(self.__problem_report)
                    return False, None
                else:
                    await self.log(
                        progress=100, message=f'Terminated with exception',
                        exception=str(e)
                    )
                    raise
            else:
                return True, ledger

    async def accept_microledger(self, leader: Pairwise, propose: InitRequestLedgerMessage) -> (bool, AbstractMicroledger):
        if self.me.did not in propose.participants:
            raise SiriusContextError('Invalid state machine initialization')
        time_to_live = propose.timeout_sec or self.time_to_live
        await self._bootstrap(propose.participants)
        async with self.leader(their=leader, thread_id=propose.thread_id, time_to_live=time_to_live) as co:
            ledger_name = propose.ledger.get('name', None)
            try:
                if not ledger_name:
                    raise StateMachineTerminatedWithError(
                        problem_code=REQUEST_PROCESSING_ERROR,
                        explain='Ledger name is Empty!',
                    )
                for their_did in propose.participants:
                    if their_did != self.me.did:
                        pw = self.__cached_p2p.get(their_did, None)
                        if pw is None:
                            raise StateMachineTerminatedWithError(
                                problem_code=REQUEST_PROCESSING_ERROR,
                                explain=f'Pairwise for DID: "their_did" does not exists!' % their_did
                            )
                await self.log(progress=0, message=f'Start ledger [{ledger_name}] creation process')
                ledger = await self._accept_microledger_internal(co, leader, propose, time_to_live)
                await self.log(progress=100, message='Ledger creation terminated successfully')
            except Exception as e:
                await sirius_sdk.Microledgers.reset(ledger_name)
                await self.log(message=f'Reset ledger')
                if isinstance(e, StateMachineTerminatedWithError):
                    self.__problem_report = SimpleConsensusProblemReport(e.problem_code, e.explain)
                    await self.log(
                        progress=100, message=f'Terminated with error',
                        problem_code=e.problem_code, explain=e.explain
                    )
                    if e.notify:
                        await co.send(self.__problem_report)
                    return False, None
                else:
                    await self.log(
                        progress=100, message=f'Terminated with exception',
                        exception=str(e)
                    )
                    raise
            else:
                return True, ledger

    async def commit(
            self, ledger: AbstractMicroledger, participants: List[str], transactions: List[Transaction]
    ) -> (bool, Optional[List[Transaction]]):
        """
        :param ledger: Microledger instance to operate with
        :param participants: list of DIDs that present pairwise list of the Microledger relationships
                (Assumed DIDs are public or every participant has relationship with each other via pairwise)
        :param transactions: transactions to commit
        """
        await self._bootstrap(participants)
        relationships = [p2p for p2p in self.__cached_p2p.values()]
        async with self.acceptors(theirs=relationships, thread_id='simple-consensus-commit-' + uuid.uuid4().hex) as co:
            try:
                await self.log(progress=0, message=f'Start committing {len(transactions)} transactions')
                txns = await self._commit_internal(co, ledger, transactions, participants)
                await self.log(progress=100, message='Commit operation was accepted by all participants')
                return True, txns
            except Exception as e:
                await ledger.reset_uncommitted()
                await self.log(message='Reset uncommitted')
                if isinstance(e, StateMachineTerminatedWithError):
                    self.__problem_report = SimpleConsensusProblemReport(e.problem_code, e.explain)
                    await self.log(
                        progress=100, message=f'Terminated with error',
                        problem_code=e.problem_code, explain=e.explain
                    )
                    if e.notify:
                        await co.send(self.__problem_report)
                    return False, None
                else:
                    await self.log(
                        progress=100, message=f'Terminated with exception',
                        exception=str(e)
                    )
                    raise

    async def commit_in_parallel(
            self, ledgers: List[Union[AbstractMicroledger, str]],
            participants: List[str], transactions: List[Transaction]
    ) -> bool:
        """
        :param ledgers: list of Microledgers instance to operate with
        :param participants: list of DIDs that present pairwise list of the Microledger relationships
                (Assumed DIDs are public or every participant has relationship with each other via pairwise)
        :param transactions: transactions to commit
        """
        await self._bootstrap(participants)
        relationships = [p2p for p2p in self.__cached_p2p.values()]
        async with self.acceptors(theirs=relationships, thread_id='simple-consensus-commit-parallel-' + uuid.uuid4().hex) as co:
            batching_api = await sirius_sdk.Microledgers.batched()
            ledgers_for_commit: List[AbstractMicroledger] = []
            if batching_api is None:
                for ledger in ledgers:
                    if isinstance(ledger, AbstractMicroledger):
                        ledgers_for_commit.append(ledger)
                    elif isinstance(ledger, str):
                        inst = await sirius_sdk.Microledgers.ledger(ledger)
                        ledgers_for_commit.append(inst)
                    else:
                        raise RuntimeError('Unexpected ledger type: %s' % str(type(ledger)))
            else:
                ledgers_for_commit = await batching_api.open(ledgers)
            try:
                try:
                    await self.log(progress=0, message=f'Start parallel committing of {len(transactions)} transactions')
                    await self._commit_internal_parallel(co, ledgers_for_commit, transactions, participants, batching_api)
                    await self.log(progress=100, message='Commit operation was accepted by all participants')
                    return True
                except Exception as e:
                    if batching_api is None:
                        for ledger in ledgers_for_commit:
                            await ledger.reset_uncommitted()
                    else:
                        await batching_api.reset_uncommitted()
                    await self.log(message='Reset uncommitted')
                    if isinstance(e, StateMachineTerminatedWithError):
                        self.__problem_report = SimpleConsensusProblemReport(e.problem_code, e.explain)
                        await self.log(
                            progress=100, message=f'Terminated with error',
                            problem_code=e.problem_code, explain=e.explain
                        )
                        if e.notify:
                            await co.send(self.__problem_report)
                        return False
                    else:
                        await self.log(
                            progress=100, message=f'Terminated with exception',
                            exception=str(e)
                        )
                        raise
            finally:
                if batching_api is not None:
                    await batching_api.close()

    async def accept_commit(self, leader: Pairwise, propose: ProposeTransactionsMessage) -> bool:
        time_to_live = propose.timeout_sec or self.time_to_live
        async with self.leader(their=leader, thread_id=propose.thread_id, time_to_live=time_to_live) as co:
            ledger = None
            try:
                await self.log(progress=0, message=f'Start acception {len(propose.transactions)} transactions')
                ledger = await self._load_ledger(propose)
                await self._accept_commit_internal(co, ledger, leader, propose)
                await self.log(progress=100, message='Acception terminated successfully')
                return True

            except Exception as e:
                if ledger:
                    await ledger.reset_uncommitted()
                    await self.log(message='Reset uncommitted')
                if isinstance(e, StateMachineTerminatedWithError):
                    self.__problem_report = SimpleConsensusProblemReport(e.problem_code, e.explain)
                    await self.log(
                        progress=100, message=f'Terminated with error',
                        problem_code=e.problem_code, explain=e.explain
                    )
                    if e.notify:
                        await co.send(self.__problem_report)
                    return False
                else:
                    raise

    async def accept_commit_parallel(self, leader: Pairwise, propose: ProposeParallelTransactionsMessage) -> bool:
        time_to_live = propose.timeout_sec or self.time_to_live
        async with self.leader(their=leader, thread_id=propose.thread_id, time_to_live=time_to_live) as co:
            batching_api = await sirius_sdk.Microledgers.batched()
            if batching_api is None:
                ledgers_for_commit: List[AbstractMicroledger] = await self._load_ledgers(propose)
            else:
                ledgers_for_commit: List[AbstractMicroledger] = await batching_api.open(propose.ledgers)
            try:
                try:
                    await self.log(progress=0, message=f'Start accept {len(propose.transactions)} transactions in parallel mode')
                    await self._accept_commit_internal_parallel(co, ledgers_for_commit, leader, propose, batching_api)
                    await self.log(progress=100, message='Accept terminated successfully in parallel mode')
                    return True
                except Exception as e:
                    if batching_api is None:
                        for ledger in ledgers_for_commit:
                            await ledger.reset_uncommitted()
                    else:
                        await batching_api.reset_uncommitted()
                    await self.log(message='Reset uncommitted')
                    if isinstance(e, StateMachineTerminatedWithError):
                        self.__problem_report = SimpleConsensusProblemReport(e.problem_code, e.explain)
                        await self.log(
                            progress=100, message=f'Terminated with error',
                            problem_code=e.problem_code, explain=e.explain
                        )
                        if e.notify:
                            await co.send(self.__problem_report)
                        return False
                    else:
                        raise
            finally:
                if batching_api is not None:
                    await batching_api.close()

    async def _bootstrap(self, participants: List[str]):
        for did in participants:
            if did != self.me.did:
                if did not in self.__cached_p2p:
                    p = await sirius_sdk.PairwiseList.load_for_did(did)
                    if p is None:
                        raise SiriusValidationError(f'Unknown pairwise for DID: {did}')
                    self.__cached_p2p[did] = p

    async def _load_ledger(self, propose: ProposeTransactionsMessage) -> AbstractMicroledger:
        try:
            await self._bootstrap(propose.participants)
            propose.validate()
            if len(propose.participants) < 2:
                raise SiriusValidationError(f'Stage-1: participant count less than 2')
            if self.me.did not in propose.participants:
                raise SiriusValidationError(f'Stage-1: {self.me.did} is not participant')
            is_ledger_exists = await sirius_sdk.Microledgers.is_exists(propose.state.name)
            if not is_ledger_exists:
                raise SiriusValidationError(f'Stage-1: Ledger with name {propose.state.name} does not exists')
        except SiriusValidationError as e:
            raise StateMachineTerminatedWithError(
                problem_code=RESPONSE_NOT_ACCEPTED,
                explain=e.message
            )
        ledger = await sirius_sdk.Microledgers.ledger(propose.state.name)
        return ledger

    async def _load_ledgers(self, propose: ProposeParallelTransactionsMessage) -> List[AbstractMicroledger]:
        ledgers = []
        try:
            await self._bootstrap(propose.participants)
            propose.validate()
            if len(propose.participants) < 2:
                raise SiriusValidationError(f'Stage-1: participant count less than 2')
            if self.me.did not in propose.participants:
                raise SiriusValidationError(f'Stage-1: {self.me.did} is not participant')
            for ledger_name in propose.ledgers:
                is_ledger_exists = await sirius_sdk.Microledgers.is_exists(ledger_name)
                if is_ledger_exists:
                    ledger = await sirius_sdk.Microledgers.ledger(ledger_name)
                    ledgers.append(ledger)
                else:
                    raise SiriusValidationError(f'Stage-1: Ledger with name {ledger_name} does not exists')
        except SiriusValidationError as e:
            raise StateMachineTerminatedWithError(
                problem_code=RESPONSE_NOT_ACCEPTED,
                explain=e.message
            )
        return ledgers

    @staticmethod
    async def _append_txns(
            ledgers: List[AbstractMicroledger], transactions: List[Transaction],
            txn_time: str = None, batching_api: Optional[AbstractBatchedAPI] = None
    ) -> List[MicroLedgerState]:
        for txn in transactions:
            if txn_time:
                txn.time = txn_time
            if not txn.has_metadata():
                raise RuntimeError('Transaction must to have metadata')
            if not txn.time:
                txn.time = txn_time
        states: List[MicroLedgerState] = []
        if batching_api is None:
            for ledger in ledgers:
                pos1, pos2, new_txns = await ledger.append(transactions)
                states.append(MicroLedgerState.from_ledger(ledger))
        else:
            _ = await batching_api.append(transactions)
            states = [MicroLedgerState.from_ledger(item) for item in _]
        return states

    async def _init_microledger_internal(
            self, co: CoProtocolThreadedTheirs, ledger: AbstractMicroledger, participants: List[str], genesis: List[Transaction]
    ):
        success, busy = await Locking.acquire(names=[ledger.name], lock_timeout=self.time_to_live)
        if not success:
            raise StateMachineTerminatedWithError(
                problem_code=REQUEST_NOT_ACCEPTED,
                explain='Preparing: Ledgers [%s] are locked by other state-machine' % ','.join(busy),
                notify=False
            )
        try:
            # ============= STAGE 1: PROPOSE =================
            propose = InitRequestLedgerMessage(
                timeout_sec=self.time_to_live,
                ledger_name=ledger.name,
                genesis=genesis,
                root_hash=ledger.root_hash,
                participants=[did for did in participants]
            )
            await propose.add_signature(sirius_sdk.Crypto, self.me)
            request_commit = InitResponseLedgerMessage()
            request_commit.assign_from(propose)

            await self.log(progress=20, message='Send propose', payload=dict(propose))

            # Switch to await transaction acceptors action
            results = await co.switch(propose)
            await self.log(progress=30, message='Received responses from all acceptors')

            errored_acceptors_did = [pairwise.their.did for pairwise, (ok, _) in results.items() if not ok]
            if errored_acceptors_did:
                raise StateMachineTerminatedWithError(
                    problem_code=REQUEST_PROCESSING_ERROR,
                    explain='Stage-1: Participants [%s] unreachable' % ','.join(errored_acceptors_did),
                )

            await self.log(progress=40, message='Validate responses')
            for pairwise, (_, response) in results.items():
                if isinstance(response, InitResponseLedgerMessage):
                    response.validate()
                    await response.check_signatures(sirius_sdk.Crypto, pairwise.their.did)
                    signature = response.signature(pairwise.their.did)
                    request_commit.signatures.append(signature)
                elif isinstance(response, SimpleConsensusProblemReport):
                    raise StateMachineTerminatedWithError(response.problem_code, response.explain)

            # ============= STAGE 2: COMMIT ============
            await self.log(progress=60, message='Send commit request', payload=dict(request_commit))
            results = await co.switch(request_commit)
            await self.log(progress=70, message='Received commit responses')
            errored_acceptors_did = [pairwise.their.did for pairwise, (ok, _) in results.items() if not ok]
            if errored_acceptors_did:
                raise StateMachineTerminatedWithError(
                    problem_code=REQUEST_PROCESSING_ERROR,
                    explain='Stage-2: Participants [%s] unreachable' % ','.join(errored_acceptors_did),
                )

            await self.log(progress=80, message='Validate commit responses from acceptors')
            for pairwise, (_, response) in results.items():
                if isinstance(response, SimpleConsensusProblemReport):
                    raise StateMachineTerminatedWithError(
                        problem_code=RESPONSE_PROCESSING_ERROR,
                        explain=f'Participant DID: {pairwise.their.did} declined operation with error: "{response.explain}"'
                    )

            # ============== STAGE 3: POST-COMMIT ============
            ack = Ack(status=Status.OK)
            await self.log(progress=90, message='All checks OK. Send Ack to acceptors')
            await co.send(ack)
        finally:
            await Locking.release()

    async def _accept_microledger_internal(
            self, co: CoProtocolThreadedP2P, leader: Pairwise, propose: InitRequestLedgerMessage, timeout: int
    ) -> AbstractMicroledger:
        success, busy = await Locking.acquire(names=[propose.ledger['name']], lock_timeout=self.time_to_live)
        if not success:
            raise StateMachineTerminatedWithError(
                problem_code=REQUEST_NOT_ACCEPTED,
                explain='Preparing: Ledgers [%s] are locked by other state-machine' % ','.join(busy),
            )
        try:
            # =============== STAGE 1: PROPOSE ===============
            try:
                propose.validate()
                await propose.check_signatures(sirius_sdk.Crypto, leader.their.did)
                if len(propose.participants) < 2:
                    raise SiriusValidationError('Stage-1: participants less than 2')
            except SiriusValidationError as e:
                raise StateMachineTerminatedWithError(
                    REQUEST_NOT_ACCEPTED, e.message
                )
            genesis = [Transaction(txn) for txn in propose.ledger['genesis']]
            await self.log(progress=10, message='Initialize ledger')
            ledger, txns = await sirius_sdk.Microledgers.create(propose.ledger['name'], genesis)
            await self.log(progress=20, message='Ledger initialized successfully')
            if propose.ledger['root_hash'] != ledger.root_hash:
                await sirius_sdk.Microledgers.reset(ledger.name)
                raise StateMachineTerminatedWithError(REQUEST_PROCESSING_ERROR, 'Stage-1: Non-consistent Root Hash')
            response = InitResponseLedgerMessage(timeout_sec=timeout)
            response.assign_from(propose)
            commit_ledger_hash = response.ledger_hash
            await response.add_signature(sirius_sdk.Crypto, self.me)
            # =============== STAGE 2: COMMIT ===============
            await self.log(progress=30, message='Send propose response', payload=dict(response))
            ok, request_commit = await co.switch(response)
            if ok:
                await self.log(progress=50, message='Validate request commit')
                if isinstance(request_commit, InitResponseLedgerMessage):
                    try:
                        request_commit.validate()
                        hashes = await request_commit.check_signatures(sirius_sdk.Crypto, participant='ALL')
                        for their_did, decoded in hashes.items():
                            if decoded != commit_ledger_hash:
                                raise SiriusValidationError(f'Stage-2: NonEqual Ledger hash with participant "{their_did}"')
                    except SiriusValidationError as e:
                        raise StateMachineTerminatedWithError(REQUEST_NOT_ACCEPTED, e.message)
                    commit_participants_set = set(request_commit.participants)
                    propose_participants_set = set(propose.participants)
                    signers_set = set([s['participant'] for s in request_commit.signatures])
                    if propose_participants_set != signers_set:
                        error_explain = 'Stage-2: Set of signers differs from proposed participants set'
                    elif commit_participants_set != signers_set:
                        error_explain = 'Stage-2: Set of signers differs from commit participants set'
                    else:
                        error_explain = None
                    if error_explain:
                        raise StateMachineTerminatedWithError(REQUEST_NOT_ACCEPTED, error_explain)
                    else:
                        # Accept commit
                        await self.log(progress=70, message='Send Ack')
                        ack = Ack(status=Status.OK)
                        ok, resp = await co.switch(ack)
                        # =========== STAGE-3: POST-COMMIT ===============
                        if ok:
                            await self.log(progress=90, message='Response to Ack received')
                            if isinstance(resp, Ack):
                                return ledger
                            elif isinstance(resp, SimpleConsensusProblemReport):
                                self.__problem_report = resp
                                logging.error(
                                    'Code: %s; Explain: %s' % (resp.problem_code, resp.explain))
                                raise StateMachineTerminatedWithError(
                                    self.__problem_report.problem_code, self.__problem_report.explain
                                )
                        else:
                            raise StateMachineTerminatedWithError(
                                problem_code=REQUEST_PROCESSING_ERROR,
                                explain='Stage-3: Commit accepting was terminated by timeout for actor: %s' % leader.their.did
                            )
                elif isinstance(request_commit, SimpleConsensusProblemReport):
                    self.__problem_report = request_commit
                    raise StateMachineTerminatedWithError(
                        self.__problem_report.problem_code, self.__problem_report.explain
                    )
            else:
                raise StateMachineTerminatedWithError(
                    problem_code=REQUEST_PROCESSING_ERROR,
                    explain='Stage-2: Commit response awaiting was terminated by timeout for actor: %s' % leader.their.did,
                )
        finally:
            await Locking.release()

    async def _commit_internal(
            self, co: CoProtocolThreadedTheirs, ledger: AbstractMicroledger, transactions: List[Transaction], participants: List[str]
    ) -> List[Transaction]:

        success, busy = await Locking.acquire(names=[ledger.name], lock_timeout=self.time_to_live)
        if not success:
            raise StateMachineTerminatedWithError(
                problem_code=REQUEST_NOT_ACCEPTED,
                explain='Preparing: Ledgers [%s] are locked by other state-machine' % ','.join(busy),
                notify=False
            )
        try:
            txn_time = str(datetime.utcnow())
            start, end, txns = await ledger.append(transactions, txn_time)
            propose = ProposeTransactionsMessage(
                transactions=txns,
                state=MicroLedgerState.from_ledger(ledger),
                participants=participants,
                timeout_sec=self.time_to_live
            )
            # ==== STAGE-1 Propose transactions to participants ====
            commit = CommitTransactionsMessage(participants=participants)
            self_pre_commit = PreCommitTransactionsMessage(state=propose.state)
            await self_pre_commit.sign_state(sirius_sdk.Crypto, self.me)
            commit.add_pre_commit(
                participant=self.me.did,
                pre_commit=self_pre_commit
            )

            await self.log(progress=20, message='Send Propose to participants', payload=dict(propose))
            results = await co.switch(propose)
            await self.log(progress=30, message='Received Propose from participants')

            errored_acceptors_did = [pairwise.their.did for pairwise, (ok, _) in results.items() if not ok]
            if errored_acceptors_did:
                raise StateMachineTerminatedWithError(
                    problem_code=REQUEST_PROCESSING_ERROR,
                    explain='Stage-1: Participants [%s] unreachable' % ','.join(errored_acceptors_did),
                )

            await self.log(progress=50, message='Validate responses')
            for pairwise, (_, pre_commit) in results.items():
                if isinstance(pre_commit, PreCommitTransactionsMessage):
                    try:
                        pre_commit.validate()
                        success, state = await pre_commit.verify_state(sirius_sdk.Crypto, pairwise.their.verkey)
                        if not success:
                            raise SiriusValidationError(
                                f'Stage-1: Error verifying signed ledger state for participant {pairwise.their.did}'
                            )
                        if pre_commit.hash != propose.state.hash:
                            raise SiriusValidationError(
                                f'Stage-1: Non-consistent ledger state for participant {pairwise.their.did}'
                            )
                    except SiriusValidationError as e:
                        raise StateMachineTerminatedWithError(
                            problem_code=RESPONSE_NOT_ACCEPTED,
                            explain=f'Stage-1: Error for participant {pairwise.their.did}: "{e.message}"'
                        )
                    else:
                        commit.add_pre_commit(pairwise.their.did, pre_commit)
                elif isinstance(pre_commit, SimpleConsensusProblemReport):
                    explain = f'Stage-1: Problem report from participant {pairwise.their.did} "{pre_commit.explain}"'
                    raise StateMachineTerminatedWithError(
                        problem_code=RESPONSE_NOT_ACCEPTED,
                        explain=explain
                    )

            # ===== STAGE-2: Accumulate pre-commits and send commit propose to all participants
            post_commit_all = PostCommitTransactionsMessage()
            await post_commit_all.add_commit_sign(sirius_sdk.Crypto, commit, self.me)

            await self.log(progress=60, message='Send Commit to participants', payload=dict(commit))
            results = await co.switch(commit)
            await self.log(progress=70, message='Received Commit response from participants')

            errored_acceptors_did = [pairwise.their.did for pairwise, (ok, _) in results.items() if not ok]
            if errored_acceptors_did:
                raise StateMachineTerminatedWithError(
                    problem_code=REQUEST_PROCESSING_ERROR,
                    explain='Stage-2: Participants [%s] unreachable' % ','.join(errored_acceptors_did),
                )

            await self.log(progress=80, message='Validate responses')
            for pairwise, (_, post_commit) in results.items():
                if isinstance(post_commit, PostCommitTransactionsMessage):
                    try:
                        post_commit.validate()
                    except SiriusValidationError as e:
                        raise StateMachineTerminatedWithError(
                            problem_code=RESPONSE_NOT_ACCEPTED,
                            explain=f'Stage-2: Error for participant {pairwise.their.did}: "{e.message}"',
                        )
                    else:
                        post_commit_all['commits'].extend(post_commit.commits)
                elif isinstance(post_commit, SimpleConsensusProblemReport):
                    raise StateMachineTerminatedWithError(
                        problem_code=RESPONSE_NOT_ACCEPTED,
                        explain=f'Stage-2: Problem report from participant {pairwise.their.did} "{post_commit.explain}"'
                    )

            # ===== STAGE-3: Notify all participants with post-commits and finalize process
            await self.log(progress=90, message='Send Post-Commit', payload=dict(post_commit_all))
            await co.send(post_commit_all)
            uncommitted_size = ledger.uncommitted_size - ledger.size
            await ledger.commit(uncommitted_size)
            return txns
        finally:
            await Locking.release()

    async def _commit_internal_parallel(
            self, co: CoProtocolThreadedTheirs, ledgers: List[AbstractMicroledger],
            transactions: List[Transaction], participants: List[str], batching_api: AbstractBatchedAPI = None
    ):
        success, busy = await Locking.acquire(names=[ledger.name for ledger in ledgers], lock_timeout=self.time_to_live)
        if not success:
            raise StateMachineTerminatedWithError(
                problem_code=REQUEST_NOT_ACCEPTED,
                explain='Preparing: Ledgers [%s] are locked by other state-machine' % ','.join(busy),
                notify=False
            )
        try:
            txn_time = str(datetime.utcnow())
            new_transactions = [Transaction.create(copy.deepcopy(txn)) for txn in transactions]  # copy original txn collection
            states = await self._append_txns(ledgers, new_transactions, txn_time, batching_api)
            propose = ProposeParallelTransactionsMessage(
                transactions=new_transactions,
                states=states,
                participants=participants,
                timeout_sec=self.time_to_live
            )
            # ==== STAGE-1 Propose transactions to participants ====
            commit = CommitParallelTransactionsMessage(participants=participants)
            self_pre_commit = PreCommitParallelTransactionsMessage(transactions=new_transactions, states=states)
            await self_pre_commit.sign_states(sirius_sdk.Crypto, self.me)
            commit.add_pre_commit(
                participant=self.me.did,
                pre_commit=self_pre_commit
            )

            await self.log(progress=20, message='Send Propose to participants (parallel mode)', payload=dict(propose))
            results = await co.switch(propose)
            await self.log(progress=30, message='Received Propose from participants (parallel mode)')

            errored_acceptors_did = [pairwise.their.did for pairwise, (ok, _) in results.items() if not ok]
            if errored_acceptors_did:
                raise StateMachineTerminatedWithError(
                    problem_code=REQUEST_PROCESSING_ERROR,
                    explain='Stage-1: Participants [%s] unreachable (parallel mode)' % ','.join(errored_acceptors_did),
                )

            await self.log(progress=50, message='Validate responses (parallel mode)')

            for pairwise, (_, pre_commit) in results.items():
                if isinstance(pre_commit, PreCommitParallelTransactionsMessage):
                    try:
                        pre_commit.validate()
                        success, state = await pre_commit.verify_state(sirius_sdk.Crypto, pairwise.their.verkey)
                        if not success:
                            raise SiriusValidationError(
                                f'Stage-1: Error verifying signed ledger state for participant {pairwise.their.did} (parallel mode)'
                            )
                        if pre_commit.hash != propose.hash:
                            raise SiriusValidationError(
                                f'Stage-1: Non-consistent ledger state for participant {pairwise.their.did} (parallel mode)'
                            )
                    except SiriusValidationError as e:
                        raise StateMachineTerminatedWithError(
                            problem_code=RESPONSE_NOT_ACCEPTED,
                            explain=f'Stage-1: Error for participant {pairwise.their.did}: "{e.message}" (parallel mode)'
                        )
                    else:
                        commit.add_pre_commit(pairwise.their.did, pre_commit)
                elif isinstance(pre_commit, SimpleConsensusProblemReport):
                    explain = f'Stage-1: Problem report from participant {pairwise.their.did} "{pre_commit.explain}" (parallel mode)'
                    raise StateMachineTerminatedWithError(
                        problem_code=RESPONSE_NOT_ACCEPTED,
                        explain=explain
                    )
            # ===== STAGE-2: Accumulate pre-commits and send commit propose to all participants
            post_commit_all = PostCommitParallelTransactionsMessage()
            await post_commit_all.add_commit_sign(sirius_sdk.Crypto, commit, self.me)

            await self.log(progress=60, message='Send Commit to participants (parallel mode)', payload=dict(commit))
            results = await co.switch(commit)
            await self.log(progress=70, message='Received Commit response from participants (parallel mode)')

            errored_acceptors_did = [pairwise.their.did for pairwise, (ok, _) in results.items() if not ok]
            if errored_acceptors_did:
                raise StateMachineTerminatedWithError(
                    problem_code=REQUEST_PROCESSING_ERROR,
                    explain='Stage-2: Participants [%s] unreachable (parallel mode)' % ','.join(errored_acceptors_did),
                )

            await self.log(progress=80, message='Validate responses (parallel mode)')
            for pairwise, (_, post_commit) in results.items():
                if isinstance(post_commit, PostCommitParallelTransactionsMessage):
                    try:
                        post_commit.validate()
                    except SiriusValidationError as e:
                        raise StateMachineTerminatedWithError(
                            problem_code=RESPONSE_NOT_ACCEPTED,
                            explain=f'Stage-2: Error for participant {pairwise.their.did}: "{e.message}" (parallel mode)',
                        )
                    else:
                        post_commit_all['commits'].extend(post_commit.commits)
                elif isinstance(post_commit, SimpleConsensusProblemReport):
                    raise StateMachineTerminatedWithError(
                        problem_code=RESPONSE_NOT_ACCEPTED,
                        explain=f'Stage-2: Problem report from participant {pairwise.their.did} "{post_commit.explain}" (parallel mode)'
                    )
            # ===== STAGE-3: Notify all participants with post-commits and finalize process
            await self.log(progress=90, message='Send Post-Commit', payload=dict(post_commit_all))
            await co.send(post_commit_all)
            if batching_api is None:
                for ledger in ledgers:
                    uncommitted_size = ledger.uncommitted_size - ledger.size
                    await ledger.commit(uncommitted_size)
            else:
                await batching_api.commit()
        finally:
            await Locking.release()

    async def _accept_commit_internal(
            self, co: CoProtocolThreadedP2P, ledger: AbstractMicroledger,
            leader: Pairwise, propose: ProposeTransactionsMessage
    ):
        success, busy = await Locking.acquire(names=[ledger.name], lock_timeout=self.time_to_live)
        if not success:
            raise StateMachineTerminatedWithError(
                problem_code=REQUEST_NOT_ACCEPTED,
                explain='Preparing: Ledgers [%s] are locked by other state-machine' % ','.join(busy),
            )
        try:
            # ===== STAGE-1: Process Propose, apply transactions and response ledger state on self-side
            await ledger.append(propose.transactions)
            ledger_state = MicroLedgerState.from_ledger(ledger)
            pre_commit = PreCommitTransactionsMessage(state=MicroLedgerState.from_ledger(ledger))
            await pre_commit.sign_state(sirius_sdk.Crypto, self.me)
            await self.log(progress=10, message='Send Pre-Commit', payload=dict(pre_commit))

            ok, commit = await co.switch(pre_commit)
            if ok:
                await self.log(progress=20, message='Received Pre-Commit response', payload=dict(commit))
                if isinstance(commit, CommitTransactionsMessage):
                    # ===== STAGE-2: Process Commit request, check neighbours signatures
                    try:
                        await self.log(progress=30, message='Validate Commit')
                        if set(commit.participants) != set(propose.participants):
                            raise SiriusValidationError('Non-consistent participants')
                        commit.validate()
                        await commit.verify_pre_commits(sirius_sdk.Crypto, ledger_state)
                    except SiriusValidationError as e:
                        raise StateMachineTerminatedWithError(
                            problem_code=REQUEST_NOT_ACCEPTED,
                            explain=f'Stage-2: error for actor {leader.their.did}: "{e.message}"'
                        )
                    else:
                        # ===== STAGE-3: Process post-commit, verify participants operations
                        post_commit = PostCommitTransactionsMessage()
                        await post_commit.add_commit_sign(sirius_sdk.Crypto, commit, self.me)

                        await self.log(progress=50, message='Send Post-Commit', payload=dict(post_commit))
                        ok, post_commit_all = await co.switch(post_commit)
                        if ok:
                            await self.log(
                                progress=60, message='Received Post-Commit response', payload=dict(post_commit_all)
                            )
                            if isinstance(post_commit_all, PostCommitTransactionsMessage):
                                try:

                                    await self.log(progress=80, message='Validate response')
                                    post_commit_all.validate()

                                    verkeys = [p2p.their.verkey for p2p in self.__cached_p2p.values()]
                                    await post_commit_all.verify_commits(sirius_sdk.Crypto, commit, verkeys)

                                except SiriusValidationError as e:
                                    raise StateMachineTerminatedWithError(
                                        problem_code=REQUEST_NOT_ACCEPTED,
                                        explain=f'Stage-3: error for leader {leader.their.did}: "{e.message}"'
                                    )
                                else:
                                    uncommitted_size = ledger_state.uncommitted_size - ledger_state.size
                                    await self.log(progress=90, message='Flush transactions to Ledger storage')
                                    await ledger.commit(uncommitted_size)
                            elif isinstance(post_commit_all, SimpleConsensusProblemReport):
                                raise StateMachineTerminatedWithError(
                                    problem_code=self.__problem_report.problem_code,
                                    explain=f'Stage-3: Problem report from leader {leader.their.did}: "{post_commit_all.explain}"'
                                )
                        else:
                            raise StateMachineTerminatedWithError(
                                problem_code=REQUEST_PROCESSING_ERROR,
                                explain=f'Stage-3: Post-Commit awaiting terminated by timeout for leader: {leader.their.did}'
                            )
                elif isinstance(commit, SimpleConsensusProblemReport):
                    explain = f'Stage-1: Problem report from leader {leader.their.did}: "{commit.explain}"'
                    self.__problem_report = SimpleConsensusProblemReport(commit.problem_code, explain)
                    raise StateMachineTerminatedWithError(
                        problem_code=self.__problem_report.problem_code, explain=self.__problem_report.explain
                    )
                else:
                    raise StateMachineTerminatedWithError(
                        problem_code=REQUEST_NOT_ACCEPTED, explain='Unexpected message @type: %s' % (str(commit.type))
                    )
            else:
                raise StateMachineTerminatedWithError(
                    problem_code=REQUEST_PROCESSING_ERROR,
                    explain=f'Stage-1: Commit awaiting terminated by timeout for leader: {leader.their.did}'
                )
        finally:
            await Locking.release()

    async def _accept_commit_internal_parallel(
            self, co: CoProtocolThreadedP2P, ledgers: List[AbstractMicroledger],
            leader: Pairwise, propose: ProposeParallelTransactionsMessage, batching_api: AbstractBatchedAPI = None
    ):
        success, busy = await Locking.acquire(names=[ledger.name for ledger in ledgers], lock_timeout=self.time_to_live)
        if not success:
            raise StateMachineTerminatedWithError(
                problem_code=REQUEST_NOT_ACCEPTED,
                explain='Preparing: Ledgers [%s] are locked by other state-machine' % ','.join(busy),
            )
        try:
            # ===== STAGE-1: Process Propose, apply transactions and response ledgers states on self-side
            states = await self._append_txns(ledgers, propose.transactions, batching_api=batching_api)
            pre_commit = PreCommitParallelTransactionsMessage(transactions=propose.transactions, states=states)
            await pre_commit.sign_states(sirius_sdk.Crypto, self.me)
            await self.log(progress=10, message='Send Pre-Commit (parallel mode)', payload=dict(pre_commit))

            ok, commit = await co.switch(pre_commit)
            if ok:
                await self.log(progress=20, message='Received Pre-Commit response (parallel mode)', payload=dict(commit))
                if isinstance(commit, CommitParallelTransactionsMessage):
                    # ===== STAGE-2: Process Commit request, check neighbours signatures
                    try:
                        await self.log(progress=30, message='Validate Commit (parallel mode)')
                        if set(commit.participants) != set(propose.participants):
                            raise SiriusValidationError('Non-consistent participants (parallel mode)')
                        commit.validate()
                        await commit.verify_pre_commits(sirius_sdk.Crypto, pre_commit.hash)
                    except SiriusValidationError as e:
                        raise StateMachineTerminatedWithError(
                            problem_code=REQUEST_NOT_ACCEPTED,
                            explain=f'Stage-2: error for actor {leader.their.did}: "{e.message}" (parallel mode)'
                        )
                    else:
                        # ===== STAGE-3: Process post-commit, verify participants operations
                        post_commit = PostCommitParallelTransactionsMessage()
                        await post_commit.add_commit_sign(sirius_sdk.Crypto, commit, self.me)

                        await self.log(progress=50, message='Send Post-Commit (parallel mode)', payload=dict(post_commit))
                        ok, post_commit_all = await co.switch(post_commit)
                        if ok:
                            await self.log(
                                progress=60, message='Received Post-Commit response (parallel mode)', payload=dict(post_commit_all)
                            )
                            if isinstance(post_commit_all, PostCommitParallelTransactionsMessage):
                                try:

                                    await self.log(progress=80, message='Validate response (parallel mode)')
                                    post_commit_all.validate()

                                    verkeys = [p2p.their.verkey for p2p in self.__cached_p2p.values()]
                                    await post_commit_all.verify_commits(sirius_sdk.Crypto, commit, verkeys)

                                except SiriusValidationError as e:
                                    raise StateMachineTerminatedWithError(
                                        problem_code=REQUEST_NOT_ACCEPTED,
                                        explain=f'Stage-3: error for leader {leader.their.did}: "{e.message}" (parallel mode)'
                                    )
                                else:
                                    if batching_api is None:
                                        for ledger in ledgers:
                                            uncommitted_size = ledger.uncommitted_size - ledger.size
                                            await ledger.commit(uncommitted_size)
                                    else:
                                        await batching_api.commit()
                                    await self.log(progress=90, message='Flush transactions to Ledger storage')
                            elif isinstance(post_commit_all, SimpleConsensusProblemReport):
                                raise StateMachineTerminatedWithError(
                                    problem_code=self.__problem_report.problem_code,
                                    explain=f'Stage-3: Problem report from leader {leader.their.did}: "{post_commit_all.explain}" (parallel mode)'
                                )
                        else:
                            raise StateMachineTerminatedWithError(
                                problem_code=REQUEST_PROCESSING_ERROR,
                                explain=f'Stage-3: Post-Commit awaiting terminated by timeout for leader: {leader.their.did} (parallel mode)'
                            )
                elif isinstance(commit, SimpleConsensusProblemReport):
                    explain = f'Stage-1: Problem report from leader {leader.their.did}: "{commit.explain}" (parallel mode)'
                    self.__problem_report = SimpleConsensusProblemReport(commit.problem_code, explain)
                    raise StateMachineTerminatedWithError(
                        problem_code=self.__problem_report.problem_code, explain=self.__problem_report.explain
                    )
                else:
                    raise StateMachineTerminatedWithError(
                        problem_code=REQUEST_NOT_ACCEPTED, explain='Unexpected message @type: %s (parallel mode)' % (str(commit.type))
                    )
            else:
                raise StateMachineTerminatedWithError(
                    problem_code=REQUEST_PROCESSING_ERROR,
                    explain=f'Stage-1: Commit awaiting terminated by timeout for leader: {leader.their.did} (parallel mode)'
                )
        finally:
            await Locking.release()



import uuid
import asyncio

import pytest

import sirius_sdk
from sirius_sdk.abstract.listener import AbstractListener
from sirius_sdk.errors.exceptions import SiriusTimeoutIO, OperationAbortedManually
from sirius_sdk.abstract.api import APIRouter, APITransport
from sirius_sdk.hub.defaults.default_apis import APIDefault
from sirius_sdk.hub.defaults.default_crypto import DefaultCryptoService
from sirius_sdk.hub.mediator import Mediator
from sirius_sdk.messaging import Message
from sirius_sdk.agent.aries_rfc.mixins import ThreadMixin
from tests.conftest import create_mediator_instance
from tests.helpers import LocalCryptoManager, LocalDIDManager, run_coroutines


@pytest.mark.asyncio
async def test_open_connection(mediator_invitation: dict):
    async with sirius_sdk.context(crypto=LocalCryptoManager(), did=LocalDIDManager()):
        my_vk = await sirius_sdk.Crypto.create_key()
        mediator = create_mediator_instance(mediator_invitation, my_vk)
        await mediator.connect()
        try:
            assert mediator.did_doc is not None
            assert mediator.me.verkey == my_vk
            assert len(mediator.endpoints) == 1
            e = mediator.endpoints[0]
            assert e.routing_keys == []
            assert e.address is not None
            assert e.is_default is True
        finally:
            await mediator.disconnect()


@pytest.mark.asyncio
async def test_routing_keys(mediator_invitation: dict):
    async with sirius_sdk.context(crypto=LocalCryptoManager(), did=LocalDIDManager()):
        my_vk = await sirius_sdk.Crypto.create_key()
        routing_vk = await sirius_sdk.Crypto.create_key()
        mediator = create_mediator_instance(mediator_invitation, my_vk, routing_keys=[routing_vk])
        await mediator.connect()
        try:
            assert len(mediator.endpoints) == 1
            e = mediator.endpoints[0]
            assert len(e.routing_keys) > 0
            assert routing_vk in str(e.routing_keys)
            assert e.address is not None
            assert e.is_default is True
        finally:
            await mediator.disconnect()


@pytest.mark.asyncio
async def test_bus(mediator_invitation: dict):
    """Check bus operations:
      - subscribe
      - publish
      - unsubscribe

    """

    thid1 = 'thread-' + uuid.uuid4().hex
    thid2 = 'thread-' + uuid.uuid4().hex
    content1 = b'Message-1'
    content2 = b'Message-2'

    async with sirius_sdk.context(crypto=LocalCryptoManager(), did=LocalDIDManager()):
        my_vk = await sirius_sdk.Crypto.create_key()
        session1 = create_mediator_instance(mediator_invitation, my_vk)
        await session1.connect()
        try:
            session2 = create_mediator_instance(mediator_invitation, my_vk)
            await session2.connect()
            try:
                # Subscribe from session-2
                for thid in [thid1, thid2]:
                    ok = await session2.bus.subscribe(thid)
                    assert ok is True
                # Publish from session-1
                for thid, content in [(thid1, content1), (thid2, content2)]:
                    num = await session1.bus.publish(thid, content)
                    assert num > 0
                # Retrieve from session-2
                for n in range(2):
                    event = await session2.bus.get_event(timeout=3)
                    assert event.payload in [content1, content2]
                # Unsubscribe from thread-2
                await session2.bus.unsubscribe(thid1)
                await asyncio.sleep(1)
                # Publish again
                for thid, num_expected in [(thid1, 0), (thid2, 1)]:
                    num = await session1.bus.publish(thid, content)
                    assert num == num_expected
                # Retrieve from session-2
                event = await session2.bus.get_event(timeout=3)
                assert event.payload == content2
                with pytest.raises(SiriusTimeoutIO):
                    await session2.bus.get_event(timeout=3)
            finally:
                await session2.disconnect()
        finally:
            await session1.disconnect()


@pytest.mark.asyncio
async def test_multiple_bus_instance_for_single_connection(mediator_invitation: dict):
    thid = 'thread-' + uuid.uuid4().hex
    content = b'Message-X'

    async with sirius_sdk.context(crypto=LocalCryptoManager(), did=LocalDIDManager()):
        my_vk = await sirius_sdk.Crypto.create_key()
        session = create_mediator_instance(mediator_invitation, my_vk)
        await session.connect()
        try:
            bus1 = await session.spawn_coprotocol()
            bus2 = await session.spawn_coprotocol()
            await bus1.subscribe(thid)
            await bus1.publish(thid, content)
            with pytest.raises(SiriusTimeoutIO):
                await bus2.get_event(timeout=3)
        finally:
            await session.disconnect()


@pytest.mark.asyncio
async def test_bus_abort(mediator_invitation: dict):
    async with sirius_sdk.context(crypto=LocalCryptoManager(), did=LocalDIDManager()):
        my_vk = await sirius_sdk.Crypto.create_key()
        session = create_mediator_instance(mediator_invitation, my_vk)
        await session.connect()
        try:
            thid = 'thread-id-' + uuid.uuid4().hex
            ok = await session.bus.subscribe(thid)
            assert ok is True

            async def __abort():
                await asyncio.sleep(1)
                await session.bus.abort()

            asyncio.ensure_future(__abort())
            with pytest.raises(OperationAbortedManually):
                await session.bus.get_event(timeout=5)
        finally:
            await session.disconnect()


@pytest.mark.asyncio
async def test_router_interface_for_wired_messages(mediator_invitation: dict):
    async with sirius_sdk.context(crypto=LocalCryptoManager(), did=LocalDIDManager()):
        my_vk = await sirius_sdk.Crypto.create_key()
        my_vk_for_messaging = await sirius_sdk.Crypto.create_key()
        mediator = create_mediator_instance(mediator_invitation, my_vk)
        await mediator.connect()
        try:
            my_router: APIRouter = mediator
            their_crypto = DefaultCryptoService()
            their_transport: APITransport = APIDefault(their_crypto)
            their_vk = await their_crypto.create_key()

            my_endpoints = await my_router.get_endpoints()
            assert len(my_endpoints) > 0
            my_endpoint = my_endpoints[0]

            listener = await my_router.subscribe()
            await asyncio.sleep(1)
            message = Message({
                '@id': 'trust-ping-message-' + uuid.uuid4().hex,
                '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/trust_ping/1.0/ping',
                "comment": "Hi. Are you listening?",
            })

            # Send packed message
            await their_transport.send(
                message=message, their_vk=my_vk_for_messaging,
                endpoint=my_endpoint.address, my_vk=their_vk, routing_keys=my_endpoint.routing_keys
            )
            print('#1')
            event = await listener.get_one(timeout=5)
            print('#2')
            assert event.message == message
            assert event.sender_verkey == their_vk
            assert event.recipient_verkey == my_vk_for_messaging
        finally:
            await mediator.disconnect()


@pytest.mark.asyncio
async def test_router_interface_for_json_messages(mediator_invitation: dict):
    async with sirius_sdk.context(crypto=LocalCryptoManager(), did=LocalDIDManager()):
        my_vk = await sirius_sdk.Crypto.create_key()
        mediator = create_mediator_instance(mediator_invitation, my_vk)
        await mediator.connect()
        try:
            my_router: APIRouter = mediator
            their_crypto = DefaultCryptoService()
            their_transport: APITransport = APIDefault(their_crypto)

            my_endpoints = await my_router.get_endpoints()
            assert len(my_endpoints) > 0
            my_endpoint = my_endpoints[0]

            listener = await my_router.subscribe()
            await asyncio.sleep(1)
            message = Message({
                '@id': 'trust-ping-message-' + uuid.uuid4().hex,
                '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/trust_ping/1.0/ping',
                "comment": "Hi. Are you listening?",
            })

            # Send packed message
            await their_transport.send(
                message=message, their_vk=[],
                endpoint=my_endpoint.address, my_vk=None, routing_keys=None
            )
            print('#1')
            event = await listener.get_one(timeout=5)
            print('#2')
            assert event.message == message
            assert event.sender_verkey is None
            assert event.recipient_verkey is None
        finally:
            await mediator.disconnect()


@pytest.mark.asyncio
async def test_router_and_bus_stream_intersection(mediator_invitation: dict):
    async with sirius_sdk.context(crypto=LocalCryptoManager(), did=LocalDIDManager()):
        my_vk = await sirius_sdk.Crypto.create_key()
        my_vk_for_messaging = await sirius_sdk.Crypto.create_key()
        session1_listener = create_mediator_instance(mediator_invitation, my_vk)
        session2_bus = create_mediator_instance(mediator_invitation, my_vk)
        await session1_listener.connect()
        await session2_bus.connect()
        try:
            their_crypto = DefaultCryptoService()
            their_transport: APITransport = APIDefault(their_crypto)
            their_vk = await their_crypto.create_key()

            my_endpoints = await session1_listener.get_endpoints()
            assert len(my_endpoints) > 0
            my_endpoint = my_endpoints[0]

            # 1. Allocate Listener
            listener = await session1_listener.subscribe()
            await asyncio.sleep(1)
            message = Message({
                '@id': 'trust-ping-message-' + uuid.uuid4().hex,
                '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/trust_ping/1.0/ping',
                "comment": "Hi. Are you listening?",
            })

            print('#')
            # Check-1: Send packed message to check listener is OK but bus listener ignore message
            await their_transport.send(
                message=message, their_vk=my_vk_for_messaging,
                endpoint=my_endpoint.address, my_vk=their_vk, routing_keys=my_endpoint.routing_keys
            )
            print('#1')
            event_listener = await listener.get_one(timeout=5)
            assert event_listener.message == message
            print('#2')
            with pytest.raises(SiriusTimeoutIO):
                await session2_bus.bus.get_event(timeout=5)
            print('#3')

            # Check-2: Send bus message and check it will be ignored by listener
            thid = 'thid-' + uuid.uuid4().hex
            await session2_bus.bus.subscribe(thid)
            await session2_bus.bus.publish(thid, b'Some-Content')
            print('#4')
            with pytest.raises(SiriusTimeoutIO):
                await listener.get_one(timeout=5)
            print('#5')
        finally:
            await session1_listener.disconnect()
            await session2_bus.disconnect()


@pytest.mark.asyncio
async def test_load_balancer_with_group_id(mediator_invitation: dict):
    group1 = 'group1'
    group2 = 'group2'
    async with sirius_sdk.context(crypto=LocalCryptoManager(), did=LocalDIDManager()):

        my_vk = await sirius_sdk.Crypto.create_key()
        my_vk_for_messaging = await sirius_sdk.Crypto.create_key()
        their_crypto = LocalCryptoManager()
        their_transport: APITransport = APIDefault(their_crypto)
        their_vk = await their_crypto.create_key()

        session1_group1 = create_mediator_instance(mediator_invitation, my_vk)
        session2_group1 = create_mediator_instance(mediator_invitation, my_vk)
        session3_group2 = create_mediator_instance(mediator_invitation, my_vk)
        await session1_group1.connect()
        await session2_group1.connect()
        await session3_group2.connect()
        try:
            my_endpoints = await session1_group1.get_endpoints()
            assert len(my_endpoints) > 0
            my_endpoint = my_endpoints[0]

            async def listen_async(session: Mediator, group: str):
                listener = await session.subscribe(group)
                event = await listener.get_one()
                return event.message, group

            fut1 = asyncio.ensure_future(listen_async(session1_group1, group1))
            fut2 = asyncio.ensure_future(listen_async(session2_group1, group1))
            fut3 = asyncio.ensure_future(listen_async(session3_group2, group2))
            await asyncio.sleep(5)

            message = Message({
                '@id': 'trust-ping-message-' + uuid.uuid4().hex,
                '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/trust_ping/1.0/ping',
                "comment": "Hi. Are you listening?",
            })

            # Send packed message
            await their_transport.send(
                message=message, their_vk=my_vk_for_messaging,
                endpoint=my_endpoint.address, my_vk=their_vk, routing_keys=my_endpoint.routing_keys
            )

            print('1')
            done, pending = await asyncio.wait(
                [
                    fut1, fut2, fut3
                ],
                timeout=5, return_when=asyncio.ALL_COMPLETED
            )
            print('#1')
            results = [tsk.result() for tsk in list(done)]
            print(repr(results))
            assert len(done) == 2
            assert len(pending) == 1
            assert all([res[0] == message for res in results])
            assert set([res[1] for res in results]) == {group1, group2}
        finally:
            await session1_group1.disconnect()
            await session2_group1.disconnect()
            await session3_group2.disconnect()


@pytest.mark.asyncio
async def test_subscribe_via_hub(mediator_uri: str, mediator_verkey: str):
    crypto1 = LocalCryptoManager()
    crypto2 = LocalCryptoManager()
    income_event = None
    income_events = []
    group_id = f'group-id-' + uuid.uuid4().hex
    msg_under_test = Message({
        '@id': 'trust-ping-message-' + uuid.uuid4().hex,
        '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/trust_ping/1.0/ping',
        "comment": "Hi. Are you listening?",
    })
    my_vk1 = await crypto1.create_key(seed='0' * 32)
    my_vk2 = await crypto2.create_key(seed='1' * 32)
    cfg1 = sirius_sdk.Config().setup_mediator(mediator_uri, my_vk1, mediator_verkey). \
        override_crypto(dependency=crypto1).override_did(dependency=LocalDIDManager())
    cfg2 = sirius_sdk.Config().setup_mediator(mediator_uri, my_vk2, mediator_verkey). \
        override_crypto(dependency=crypto2).override_did(dependency=LocalDIDManager())

    async with sirius_sdk.context(cfg1):
        endpoints = await sirius_sdk.endpoints()
        cfg1_endpoint = endpoints[0].address

    async def listener(cfg):
        nonlocal income_event
        async with sirius_sdk.context(cfg):
            lst = await sirius_sdk.subscribe(group_id)
            async for event in lst:
                income_event = event
                income_events.append(event)
            print('')

    async def sender(cfg):
        await asyncio.sleep(1)
        async with sirius_sdk.context(cfg):
            try:
                await sirius_sdk.send(msg_under_test, my_vk1, cfg1_endpoint, my_vk2)
                return 'success'
            except Exception as e:
                print('Send Err: ' + repr(e))
                raise

    results = await run_coroutines(listener(cfg1), sender(cfg2), timeout=5)
    assert len(results) == 1 and results[0] == 'success'
    assert income_event is not None
    assert income_event.message == msg_under_test


@pytest.mark.asyncio
async def test_coprotocols_ping_pong(mediator_uri: str, mediator_verkey: str):
    """Local manager provide services on recipient-side, so it should use bus to publish threads to non-trusted
       environment on mediator-side
    """
    crypto_manager1 = LocalCryptoManager()
    did_manager1 = LocalDIDManager(crypto=crypto_manager1)
    crypto_manager2 = LocalCryptoManager()
    did_manager2 = LocalDIDManager(crypto=crypto_manager2)
    my_vk1 = await crypto_manager1.create_key()
    my_vk2 = await crypto_manager2.create_key()
    cfg1 = sirius_sdk.Config().setup_mediator(mediator_uri, my_vk1, mediator_verkey). \
        override_crypto(dependency=crypto_manager1).override_did(dependency=did_manager1)
    cfg2 = sirius_sdk.Config().setup_mediator(mediator_uri, my_vk2, mediator_verkey). \
        override_crypto(dependency=crypto_manager2).override_did(dependency=did_manager2)
    thread_id = 'thread-id-' + uuid.uuid4().hex
    test_msg_count = 2

    # Build P2P
    async with sirius_sdk.context(cfg1):
        # Get endpoints
        agent1_endpoint = [e for e in await sirius_sdk.endpoints() if e.routing_keys == []][0].address
        # Init pairwise list #1
        did1, verkey1 = await sirius_sdk.DID.create_and_store_my_did()

    async with sirius_sdk.context(cfg2):
        # Get endpoints
        agent2_endpoint = [e for e in await sirius_sdk.endpoints() if e.routing_keys == []][0].address
        # Init pairwise list #1
        did2, verkey2 = await sirius_sdk.DID.create_and_store_my_did()

    async with sirius_sdk.context(cfg1):
        await sirius_sdk.DID.store_their_did(did2, verkey2)
    async with sirius_sdk.context(cfg2):
        await sirius_sdk.DID.store_their_did(did1, verkey1)

    # Init pairwise list #2
    p2p1 = sirius_sdk.Pairwise(
        me=sirius_sdk.Pairwise.Me(
            did=did1,
            verkey=verkey1
        ),
        their=sirius_sdk.Pairwise.Their(
            did=did2,
            label='Label-2',
            endpoint=agent2_endpoint,
            verkey=verkey2
        )
    )
    p2p2 = sirius_sdk.Pairwise(
        me=sirius_sdk.Pairwise.Me(
            did=did2,
            verkey=verkey2
        ),
        their=sirius_sdk.Pairwise.Their(
            did=did1,
            label='Label-1',
            endpoint=agent1_endpoint,
            verkey=verkey1
        )
    )

    async def requester(cfg: sirius_sdk.Config, to: sirius_sdk.Pairwise):
        async with sirius_sdk.context(cfg):
            print('%')
            bus = await sirius_sdk.spawn_coprotocol()
            await bus.subscribe(thid=thread_id)
            print('%')
            await asyncio.sleep(1)
            for n in range(test_msg_count):
                print('#')
                req = Message({
                    '@id': 'trust-ping-message-' + uuid.uuid4().hex,
                    '@type': 'https://didcomm.org/trust_ping/1.0/ping',
                    "comment": "Hi. Are you listening?",
                    "counter": n,
                    "done": n == test_msg_count-1
                })
                ThreadMixin.set_thread(req, ThreadMixin.Thread(thid=thread_id))
                print('#')
                await sirius_sdk.send_to(req, to)
                print('#')
                resp = await bus.get_message()
                print('#')
        print('Done')

    async def responder(cfg: sirius_sdk.Config, to: sirius_sdk.Pairwise):
        async with sirius_sdk.context(cfg):
            print('%')
            bus = await sirius_sdk.spawn_coprotocol()
            await bus.subscribe(thid=thread_id)
            print('%')
            await asyncio.sleep(1)
            while True:
                print('#')
                event = await bus.get_message()
                req = event.message
                assert str(req.type) == 'https://didcomm.org/trust_ping/1.0/ping'
                resp = Message({
                    '@id': 'trust-ping-message-' + uuid.uuid4().hex,
                    '@type': 'https://didcomm.org/trust_ping/1.0/ping_response',
                    "counter": req.get('counter')
                })
                ThreadMixin.set_thread(resp, ThreadMixin.Thread(thid=thread_id))
                print('#')
                await sirius_sdk.send_to(resp, to)
                print('#')
                if req['done'] is True:
                    return

    async def decrypting_in_background(cfg: sirius_sdk.Config, label: str = ''):
        async with sirius_sdk.context(cfg):
            print(f'{label}')
            listener = await sirius_sdk.subscribe(f'DECRYPTING-{label}')
            bus = await sirius_sdk.spawn_coprotocol()
            async for e in listener:
                if e.message and isinstance(e.message, Message) and e.jwe:
                    thread = ThreadMixin.get_thread(e.message)
                    if thread and thread.thid:
                        num = await bus.publish(thread.thid, e.jwe)
                        assert num > 0

    await run_coroutines(
        requester(cfg1, p2p1),
        responder(cfg2, p2p2),
        decrypting_in_background(cfg1, 'Requester'),
        decrypting_in_background(cfg2, 'Responder'),
        timeout=5
    )



"""
The Queue Transport is a special form of transport where messages are held at the sender for pickup by the recipient.
This is useful in conditions where the recipient does not have a reliable endpoint available for message reception.

Details: https://github.com/decentralized-identity/didcomm-messaging/blob/master/extensions/return_route/main.md#queue-transport
"""
URI_QUEUE_TRANSPORT = 'didcomm:transport/queue'



import json
from enum import Enum
from dataclasses import dataclass
from typing import Union, List, Optional

from sirius_sdk.encryption import b64_to_bytes

from .errors import EncryptionError


class ConfidentialStorageEncType(Enum):
    # This enc-type typically used to save chunked structure of stream
    # that was encoded outside (on upper levels)
    UNKNOWN = 'UNKNOWN'
    # X25519
    X25519KeyAgreementKey2019 = 'X25519KeyAgreementKey2019'


@dataclass
class KeyPair:
    # Public Key
    pk: str

    # SecretKey
    sk: str
    

@dataclass
class EncHeader:
    kid: str
    sender: str = None
    iv: str = None


@dataclass
class EncRecipient:
    encrypted_key: str
    header: EncHeader


@dataclass
class JWE:
    enc: str = 'xchacha20poly1305_ietf'
    typ: str = 'JWE/1.0'
    alg: str = 'Anoncrypt'
    iv: str = None
    recipients: List[EncRecipient] = None

    def __post_init__(self):
        if self.recipients is None:
            self.recipients = []

    def as_json(self) -> dict:
        return {
            'enc': self.enc,
            'typ': self.typ,
            'alg': self.alg,
            'iv': self.iv,
            'recipients': [
                {
                    'encrypted_key': recip.encrypted_key,
                    'header': {
                        'kid': recip.header.kid,
                        'sender': recip.header.sender,
                        'iv': recip.header.iv
                    }
                }
                for recip in self.recipients
            ]
        }

    def from_json(self, js: dict):
        try:
            self.enc = js['enc']
            self.typ = js['typ']
            self.alg = js['alg']
            self.iv = js.get('iv', None)
            self.recipients = [
                EncRecipient(
                    encrypted_key=recip['encrypted_key'],
                    header=EncHeader(
                        kid=recip['header']['kid'],
                        sender=recip['header'].get('sender', None),
                        iv=recip['header'].get('iv', None)
                    )
                )
                for recip in js['recipients']
            ]
        except (LookupError, ValueError) as e:
            msg = str(e)
            raise EncryptionError(msg)


def parse_protected(jwe: Union[str, bytes, dict]) -> dict:
    if isinstance(jwe, str):
        d = json.loads(jwe)
    elif isinstance(jwe, bytes):
        d = json.loads(jwe.decode())
    else:
        d = jwe
    jwe_json = json.loads(b64_to_bytes(d["protected"], urlsafe=True).decode("ascii"))
    return jwe_json



import asyncio
import uuid

import pytest

from sirius_sdk.hub.backgrounds import BackgroundScheduler


@pytest.mark.asyncio
async def test_sane_for_empty_task_hash():

    work_count = 5
    queue1 = asyncio.Queue(maxsize=work_count)
    queue2 = asyncio.Queue(maxsize=work_count)

    async def __task1__():
        for n in range(work_count):
            queue1.put_nowait(n)

    async def __task2__():
        for n in range(work_count):
            queue2.put_nowait(n)

    # Check for empty task_hash
    BackgroundScheduler.schedule(coro=__task1__(), task_hash=None)
    BackgroundScheduler.schedule(coro=__task2__(), task_hash=None)
    await asyncio.wait([queue1.join(), queue2.join()], timeout=5)


@pytest.mark.asyncio
async def test_sane_for_filled_task_hash():

    work_count = 5
    queue1 = asyncio.Queue()
    queue2 = asyncio.Queue()
    task_hash = 'Hash-Value-' + uuid.uuid4().hex

    async def __task1__():
        for n in range(work_count):
            await asyncio.sleep(0.1)
            queue1.put_nowait(n+10)

    async def __task2__():
        for n in range(work_count):
            await asyncio.sleep(0.1)
            queue2.put_nowait(n+30)

    # Check for empty task_hash
    BackgroundScheduler.schedule(coro=__task1__(), task_hash=task_hash)
    BackgroundScheduler.schedule(coro=__task2__(), task_hash=task_hash)
    await asyncio.sleep(3)
    assert queue1.empty() is False and queue1.qsize() == work_count
    assert queue2.empty() is True


@pytest.mark.asyncio
async def test_unschedule():

    task_queue = asyncio.Queue()
    task_hash = 'Hash-Value-' + uuid.uuid4().hex

    async def __task__():
        while True:
            await asyncio.sleep(0.1)
            task_queue.put_nowait(None)

    BackgroundScheduler.schedule(coro=__task__(), task_hash=task_hash)
    BackgroundScheduler.schedule(coro=__task__(), task_hash=task_hash)
    await asyncio.sleep(1)
    count1 = task_queue.qsize()
    assert count1 > 0

    BackgroundScheduler.unschedule(task_hash=task_hash)
    await asyncio.sleep(1)
    count2 = task_queue.qsize()
    assert count2 > count1

    BackgroundScheduler.unschedule(task_hash=task_hash)
    await asyncio.sleep(1)
    count3 = task_queue.qsize()
    assert count3 == count2




import asyncio

import aiohttp


class EndpointTransport:

    def __init__(self, keepalive_timeout: float = 15):
        self.__keep_alive_connector = aiohttp.TCPConnector(
            ssl=False,  # Do not verify SSL due to all messages encrypted
            use_dns_cache=False,
            limit=0,  # No connections limit
            keepalive_timeout=keepalive_timeout
        )
        self.__websockets = {}

    def __del__(self):
        loop = asyncio.get_event_loop()
        if self.__websockets:
            if loop:
                if loop.is_running():
                    loop.create_task(self.close())
                else:
                    loop.run_until_complete(self.close())

    async def send(
            self, msg: bytes, endpoint: str, timeout: float, content_type='application/ssi-agent-wire'
    ) -> (bool, str):
        parts = endpoint.split('://')
        if len(parts) != 2:
            raise ValueError('Expected scheme in address: %s' % endpoint)
        scheme = parts[0]
        if scheme in ['http', 'https']:
            success, body = await self.http_send(msg, endpoint, timeout, content_type)
            return success, body.decode()
        elif scheme in ['ws', 'wss']:
            await self.ws_send(msg, endpoint, timeout)
            return True, ''

    async def http_send(
            self, msg: bytes, endpoint: str, timeout: float, content_type: str = 'application/ssi-agent-wire'
    ) -> (bool, bytes):
        tm = aiohttp.ClientTimeout(total=timeout)
        headers = {'content-type': content_type}
        request = aiohttp.request(
            'post', endpoint, data=msg, headers=headers, connector=self.__keep_alive_connector, timeout=tm
        )
        async with request as resp:
            body = await resp.read()
            if resp.status in [200, 202]:
                return True, body
            else:
                return False, body

    async def ws_send(self, msg: bytes, endpoint: str, timeout: float):
        tup = self.__websockets.pop(endpoint, None)
        if tup is None:
            tm = aiohttp.ClientTimeout(total=timeout)
            session = aiohttp.ClientSession(timeout=tm)
            ws = await session.ws_connect(url=endpoint)
            self.__websockets[endpoint] = (ws, session)
        else:
            ws, session = tup
            if ws.closed:
                ws = session.ws_connect(url=endpoint)
        await ws.send_bytes(msg)
        self.__websockets[endpoint] = (ws, session)

    async def close(self):
        for ws, session in self.__websockets.values():
            await ws.close()
            await session.close()


async def http_send(
        msg: bytes, endpoint: str, timeout: float,
        connector: aiohttp.TCPConnector = None,
        content_type: str = 'application/ssi-agent-wire'
):
    """Send over HTTP"""

    headers = {'content-type': content_type}
    request = aiohttp.request(
        'post', endpoint,
        data=msg,
        headers=headers,
        connector=connector,
        timeout=aiohttp.ClientTimeout(total=timeout)
    )
    async with request as resp:
        body = await resp.read()
        if resp.status in [200, 202]:
            return True, body
        else:
            return False, body
